package livresip.uaserver;
import java.io.ByteArrayInputStream;
import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintWriter;
import java.net.InetAddress;
import java.net.Socket;
import java.net.UnknownHostException;
import java.text.DateFormat;
import java.util.Date;
import java.util.Properties;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TooManyListenersException;

import javax.sip.Dialog;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.InvalidArgumentException;
import javax.sip.ListeningPoint;
import javax.sip.PeerUnavailableException;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipException;
import javax.sip.SipFactory;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.SipStack;
import javax.sip.TimeoutEvent;
import javax.sip.Transaction;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.header.ContactHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.ToHeader;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;

import livresip.uaclient.UserAgentClient;

import org.apache.log4j.Logger;
import org.apache.log4j.PropertyConfigurator;


/**
 * 	Simple SIP User Agent Server  
 * 
 *  @see Chapitre 4 du livre Tlphonie SIP : concepts, usages et programmation en Java
 *	@author G. Vanwormhoudt
 */

public class UserAgentServer implements SipListener {

	protected static Logger logger = Logger.getLogger(UserAgentClient.class);
	
	protected static SipFactory sipFactory; 
	protected static AddressFactory addressFactory;
	protected static MessageFactory messageFactory;
	protected static HeaderFactory headerFactory;
	protected static SipStack sipStack;

	protected static SipProvider sipProvider;
	protected ListeningPoint listeningPoint;

	protected DayTimeService service;

	
	public UserAgentServer(int port, String transport) 
		throws SipException, TooManyListenersException, InvalidArgumentException {
		
		listeningPoint = sipStack.createListeningPoint(sipStack.getIPAddress(), port, transport);
		sipProvider = sipStack.createSipProvider(listeningPoint);
		sipProvider.addSipListener(this); 

		logger.info("User agent: " + sipStack.getIPAddress());
        logger.info("Now listening at " + listeningPoint.getIPAddress() + ":" + listeningPoint.getPort() + "/" + listeningPoint.getTransport());
	}

	public void processRequest(RequestEvent requestEvent) {
		Request request = requestEvent.getRequest();
		ServerTransaction serverTransaction =
			requestEvent.getServerTransaction();
			
		if (request.getMethod().equals(Request.INVITE)) {
			processInvite(request, serverTransaction);
		} else if (request.getMethod().equals(Request.ACK)) {
			processAck(request, serverTransaction);
		} else if (request.getMethod().equals(Request.BYE)) {
			processBye(request, serverTransaction);
		} else { 
			requestNotSupported(request, serverTransaction); 
		}
	}

	protected void requestNotSupported(Request request,
			ServerTransaction serverTransaction) {
		
		logger.warn("Not supported Received Request " + request.getMethod());
		
		try {
			Response response = messageFactory.createResponse(Response.NOT_IMPLEMENTED, request);
			if (serverTransaction == null) {
				serverTransaction = sipProvider.getNewServerTransaction(request);
			}
			serverTransaction.sendResponse(response);
		} catch (Exception e) {
			e.printStackTrace();
		} 
	}

	/** 
	 * Process the ACK request. 
	 */
	public void processAck(Request request, ServerTransaction serverTransaction) {
		// Dmarrage du service aprs acquittement
		new Thread(service).start();
	}

	/** 
	 * Process the invite request.
	 */
	public void processInvite(Request request, ServerTransaction serverTransaction) {
	
		try {
			Response response = messageFactory.createResponse(Response.OK, request);
			ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
			String tag = Integer.toHexString(new Random().nextInt());
			toHeader.setTag(tag);
			
			// Create contact headers
			String host = sipStack.getIPAddress();
			SipURI contactURI = addressFactory.createSipURI(toHeader.getName(), host);
			contactURI.setPort(listeningPoint.getPort());

			Address contactAddress = addressFactory.createAddress(contactURI);
			ContactHeader contactHeader =
				headerFactory.createContactHeader(contactAddress);
			response.addHeader(contactHeader);

			byte[] content = request.getRawContent();
			String contentStr = new String(content);
	        
		    Properties p = new Properties();
		    try {
		    	p.load(new ByteArrayInputStream(content));
	            String ip = p.getProperty("ip");
	            int port = Integer.parseInt(p.getProperty("port"));
	            int frequency = Integer.parseInt(p.getProperty("frequency"));
	            int duration = Integer.parseInt(p.getProperty("duration"));
	            service = new DayTimeService(ip, port, frequency, duration);
	                 
	        } catch(Exception e) { 
	        	e.printStackTrace(); 
	        }
	        
		    if (serverTransaction == null) 
		    	serverTransaction = sipProvider.getNewServerTransaction(request);
		    
		    serverTransaction.sendResponse(response);
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	/** 
	 * Process the bye request.
	 */
	public void processBye(Request request, ServerTransaction serverTransaction) {	
		try {
			service.stop();
			service = null; 
			
			Response response =	messageFactory.createResponse(200, request);
			serverTransaction.sendResponse(response);
		
		} catch (Exception ex) {
			ex.printStackTrace();
		}
	}

	public void processResponse(ResponseEvent responseReceivedEvent) {
		Response response = responseReceivedEvent.getResponse();
		logger.warn("Response received : " + response.getReasonPhrase() + "-- ignore");
	}

	public void processTimeout(TimeoutEvent event) {
		Transaction transaction = (event.isServerTransaction()) 
			? (Transaction) event.getServerTransaction() : event.getClientTransaction();
		logger.trace("Transaction Time out:" + transaction);
	}

	public void processDialogTerminated(DialogTerminatedEvent event) {
		Dialog dialog = event.getDialog();
		logger.trace("DialogTerminated:" + dialog);
	}

	public void processIOException(IOExceptionEvent event) {
		Object source = event.getSource();
		logger.trace("IOException from source:" + source);
	}

	public void processTransactionTerminated(TransactionTerminatedEvent event) {
		Transaction transaction = (event.isServerTransaction()) 
			? (Transaction) event.getServerTransaction() : event.getClientTransaction();
		logger.trace("Transaction Time out:" + transaction);
	}

	
	public static void initSip(String host) throws PeerUnavailableException  {			
		// Choix de la pile
		sipFactory = SipFactory.getInstance();
		sipFactory.setPathName("gov.nist");

		// Creation des fabriques
		headerFactory = sipFactory.createHeaderFactory();
		addressFactory = sipFactory.createAddressFactory();
		messageFactory = sipFactory.createMessageFactory();
		
		Properties properties = new Properties();
		properties.setProperty("javax.sip.STACK_NAME", "UserAgentServer");
		properties.setProperty("javax.sip.IP_ADDRESS", host);
		
		properties.setProperty("gov.nist.javax.sip.TRACE_LEVEL", "16");
		
		// Creation de la pile SIP
		sipStack = sipFactory.createSipStack(properties);
	}


	public static void main(String args[]) throws Exception {
		Properties logProperties = new Properties();
		logProperties.put("log4j.rootLogger", "TRACE,UAS");
		logProperties.put("log4j.appender.UAS", "org.apache.log4j.ConsoleAppender");  // Our custom appender
		logProperties.put("log4j.appender.UAS.layout", "org.apache.log4j.PatternLayout"); //See: http://logging.apache.org/log4j/docs/api/org/apache/log4j/PatternLayout.html
		logProperties.put("log4j.appender.UAS.layout.ConversionPattern", "%d{HH:mm:ss} %5.5p %40.40c: %m%n");		
		
		PropertyConfigurator.configure(logProperties);
		
		String localIP = InetAddress.getLocalHost().getHostAddress();
		int port = 5070;
		String transport = "udp";
		
		UserAgentServer.initSip(localIP);
		UserAgentServer uas = new UserAgentServer(port, transport);
	}
	
	class DayTimeService implements Runnable {
		protected volatile boolean stop;
		protected String address; 
		protected int port;
		protected int frequency;
		protected int duration;
				
		DayTimeService(String ip, int p, int f, int d) throws 
			IOException, UnknownHostException {
			address = ip;
			duration = d;
			stop = false;
			port = p;
		}
		
		public void stop() { stop = true ;}
		
		public void run() {
			Timer stopTimer = new Timer();
			DateFormat formater = DateFormat.getDateTimeInstance(
					DateFormat.LONG, DateFormat.LONG );
			try {
				Socket socket = new Socket(InetAddress.getByName(address), port);
				OutputStream os = socket.getOutputStream();
				PrintWriter pw = new PrintWriter( os, true );
			
				stopTimer.schedule(
					new TimerTask() { public void run()	{ stop = true; }},
					duration);
			
				while (!stop) {
					String strTime = formater.format( new Date() );
					pw.println( strTime );
					try {
						Thread.sleep(frequency);
					} catch(InterruptedException e) {}
					
				}
				pw.close();
				socket.close();
			} catch(Exception e) {
				e.printStackTrace();
			}
		}		
	}

	
}
