package livresip.chat;

import java.text.ParseException;
import java.util.ArrayList;
import java.util.Properties;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TooManyListenersException;

import javax.sip.ClientTransaction;
import javax.sip.Dialog;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.InvalidArgumentException;
import javax.sip.ListeningPoint;
import javax.sip.ObjectInUseException;
import javax.sip.PeerUnavailableException;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipException;
import javax.sip.SipFactory;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.SipStack;
import javax.sip.TimeoutEvent;
import javax.sip.Transaction;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.TransportNotSupportedException;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.address.URI;
import javax.sip.header.CSeqHeader;
import javax.sip.header.CallIdHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.ContentTypeHeader;
import javax.sip.header.EventHeader;
import javax.sip.header.ExpiresHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.SubscriptionStateHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;


/**
 * 	Classe implantant la couche SIP de la messagerie conversationnelle  
 * 
 *  @see Chapitre 6 du livre Tlphonie SIP : concepts, usages et programmation en Java
 *	@author G. Vanwormhoudt
 */

public class ChatSip implements SipListener {

    protected SipStack sipStack;
    protected SipFactory sipFactory;
    protected AddressFactory addressFactory;
    protected HeaderFactory headerFactory;
    protected MessageFactory messageFactory;
    protected SipProvider sipProvider;
	protected long cSeq;
	protected ListeningPoint listeningPoint;

	protected Dialog inviteDialog;
	protected Dialog subscribeDialog;
	protected TimerTask timerTask;
	
	protected Address myAddress;
	protected Address otherPeerAddress;
    protected String username;

    protected ChatSipEventListener chatListener;

	protected boolean pendingInvite;
	protected boolean pendingSubscribe;
    	

    public ChatSip(String username, String ip, int port)
	    throws PeerUnavailableException, TransportNotSupportedException,
	    InvalidArgumentException, ObjectInUseException,
	    TooManyListenersException, ParseException {
		
    	setUsername(username);
		
		sipFactory = SipFactory.getInstance();
		sipFactory.setPathName("gov.nist");
		Properties properties = new Properties();
		properties.setProperty("javax.sip.STACK_NAME", "SipChat");
		properties.setProperty("javax.sip.IP_ADDRESS", ip);
		
		sipStack = sipFactory.createSipStack(properties);
		headerFactory = sipFactory.createHeaderFactory();
		addressFactory = sipFactory.createAddressFactory();
		messageFactory = sipFactory.createMessageFactory();
	
		listeningPoint = sipStack.createListeningPoint(ip, port, "udp");
			
		sipProvider = sipStack.createSipProvider(listeningPoint);
		sipProvider.addSipListener(this);
		
		SipURI from = addressFactory.createSipURI(username, getHost() + ":" + getPort());
		myAddress = addressFactory.createAddress(from);
		myAddress.setDisplayName(getUsername());
		
		pendingInvite = false;
    }

    
    public boolean isInChatSession() {
    	return (inviteDialog != null); 
    }
    
    public boolean isInSubscribeSession() {
    	return (subscribeDialog != null);
    }
    
    public boolean isChatSessionTransaction(Transaction transaction) {
    	return (inviteDialog == transaction.getDialog());
    }
    
    public boolean isSubscribeSessionTransaction(Transaction transaction) {
    	return (subscribeDialog == transaction.getDialog());
    }
    
 

    public void sendInvite(String to) throws ParseException, InvalidArgumentException, SipException  {
    	this.sendInvite(to, to);
    }
    
	public void sendInvite(String to, String uri)  {
		
		try {
			if (!isInChatSession()) {
				FromHeader fromHeader = headerFactory.createFromHeader(myAddress, genTag());
				
				String toUsername = to.substring(to.indexOf(":") + 1, to.indexOf("@"));
				String toDomain = to.substring(to.indexOf("@") + 1);
		
				String uriUsername = uri.substring(to.indexOf(":") + 1, to.indexOf("@"));
				String uriDomain = uri.substring(to.indexOf("@") + 1);
				
				SipURI toURI = addressFactory.createSipURI(toUsername, toDomain);
				SipURI requestURI = addressFactory.createSipURI(uriUsername, uriDomain); 
				
				Address toAddress = addressFactory.createAddress(addressFactory.createSipURI(toUsername, toDomain));
				toAddress.setDisplayName(username);
				otherPeerAddress = toAddress; 
				ToHeader toHeader = headerFactory.createToHeader(toAddress, null);
						
				// Creation d'entete Message
				CallIdHeader callIDHeader = sipProvider.getNewCallId();
				CSeqHeader cseqHeader = headerFactory.createCSeqHeader((++cSeq), Request.INVITE);
				MaxForwardsHeader maxForwards = headerFactory.createMaxForwardsHeader(70);
				
				ContactHeader contactHeader = headerFactory.createContactHeader(myAddress);
		
				// Creation d'entete Via		
				String transport = "udp";
				ArrayList<ViaHeader> viaHeaders = new ArrayList<ViaHeader>();
				ViaHeader viaHeader = headerFactory.createViaHeader(listeningPoint.getIPAddress(), listeningPoint.getPort(), transport, null);
				viaHeaders.add(viaHeader);
		
				// Creation Message
				Request invite = messageFactory.createRequest(requestURI, Request.INVITE, callIDHeader, cseqHeader, fromHeader, toHeader, viaHeaders, maxForwards);
				invite.addHeader(contactHeader);
						
				// creation de la transaction
				ClientTransaction clientTransaction = sipProvider.getNewClientTransaction(invite);
				clientTransaction.sendRequest();
				
				pendingInvite = true;
				
				chatListener.notify(EventType.SESSION_REQUESTED, null);
			}
		} catch (Exception e) {
			handleNonFatalException(e);
		} 
	}

	public void sendMessage(String message) throws ParseException,
	    InvalidArgumentException, SipException {

		if (isInChatSession()) {
			Request request = inviteDialog.createRequest(Request.MESSAGE);
			ContentTypeHeader contentTypeHeader = headerFactory.createContentTypeHeader("text", "plain");
			request.setContent(message, contentTypeHeader);
	    	
			ClientTransaction ct = sipProvider.getNewClientTransaction(request);
			inviteDialog.sendRequest(ct);
		}
   }
	
	public void sendBye() {
		try  {
			if (isInChatSession()) {
				ContactHeader contactHeader = headerFactory.createContactHeader(myAddress);
	
				// Creation Message
				Request bye = inviteDialog.createRequest(Request.BYE);
				bye.addHeader(contactHeader);
	
				// creation de la transaction
				ClientTransaction clientTransaction = sipProvider.getNewClientTransaction(bye);
			
				// Envoi Message
				inviteDialog.sendRequest(clientTransaction);
			}
		} catch(Exception e) {
			handleNonFatalException(e);
		}
	} 
	
	protected void sendSubscribe() {
		try {
			if (!isInChatSession() && !isInSubscribeSession()) {
				FromHeader fromHeader = headerFactory.createFromHeader(myAddress,genTag());
				ToHeader toHeader = headerFactory.createToHeader(otherPeerAddress, null);
					
				// Creation d'entete Message
				CallIdHeader callIDHeader = sipProvider.getNewCallId();
				CSeqHeader cseqHeader = headerFactory.createCSeqHeader((++cSeq), Request.SUBSCRIBE);
				MaxForwardsHeader maxForwards = headerFactory.createMaxForwardsHeader(70);
		
				ContactHeader contactHeader = headerFactory.createContactHeader(myAddress);
		
				// Creation entete Expires
				ExpiresHeader expiresHeader = headerFactory.createExpiresHeader(300);
				
				EventHeader eventHeader = headerFactory.createEventHeader("availability");
				
				// Creation d'entete Via		
				String transport = "udp";
				ArrayList<ViaHeader> viaHeaders = new ArrayList<ViaHeader>();
				ViaHeader viaHeader = headerFactory.createViaHeader(listeningPoint.getIPAddress(), listeningPoint.getPort(), transport, null);
				viaHeaders.add(viaHeader);
		
				// Creation Message
				URI requestURI = otherPeerAddress.getURI();
				Request subscribe = messageFactory.createRequest(requestURI, Request.SUBSCRIBE, callIDHeader, cseqHeader, fromHeader, toHeader, viaHeaders, maxForwards);
				subscribe.addHeader(contactHeader);
				subscribe.addHeader(expiresHeader);
				subscribe.addHeader(eventHeader);
				
				// creation de la transaction
				ClientTransaction clientTransaction = sipProvider.getNewClientTransaction(subscribe);
				clientTransaction.sendRequest();
		
				chatListener.notify(EventType.SUBSCRIPTION_REQUESTED, null);
			}
		} catch (Exception e) {
			handleNonFatalException(e);
		}
	}
	
	protected void sendTerminateSubscribe() {
		
		try {
			if (isInSubscribeSession()) {
				Request subscribe = subscribeDialog.createRequest(Request.SUBSCRIBE);
				ExpiresHeader expiresHeader = headerFactory.createExpiresHeader(0);		
				subscribe.setExpires(expiresHeader);
				
				ClientTransaction clientTransaction = sipProvider.getNewClientTransaction(subscribe);
				subscribeDialog.sendRequest(clientTransaction); 
			}
		} catch (Exception e) {
			handleNonFatalException(e);
		}
	}



	protected void sendActiveNotify(boolean isAvailable) {
		try {
			if (isInSubscribeSession()) {
				Request notify = subscribeDialog.createRequest(Request.NOTIFY);
				
				SubscriptionStateHeader subscriptionStateHeader = headerFactory.createSubscriptionStateHeader("active");
				notify.addHeader(subscriptionStateHeader); 
				
				EventHeader eventHeader = headerFactory.createEventHeader("");
				notify.addHeader(eventHeader);
				
				String availability = isAvailable ? "available" : "unavailable";
				ContentTypeHeader contentTypeHeader = headerFactory.createContentTypeHeader("application", "text");
				notify.setContent(availability, contentTypeHeader);
				
				ClientTransaction transaction = sipProvider.getNewClientTransaction(notify);
				subscribeDialog.sendRequest(transaction);
			}
		} catch (Exception e) {
			handleNonFatalException(e);
		}		
	}
	
	protected void sendTerminatedNotify() {
		try {
			if (isInSubscribeSession()) {
				Request notify = subscribeDialog.createRequest(Request.NOTIFY);
				
				SubscriptionStateHeader subscriptionStateHeader = headerFactory.createSubscriptionStateHeader("terminated");
				notify.addHeader(subscriptionStateHeader); 
				
				EventHeader eventHeader = headerFactory.createEventHeader("");
				notify.addHeader(eventHeader);
									
				ClientTransaction transaction = sipProvider.getNewClientTransaction(notify);
				subscribeDialog.sendRequest(transaction);
			}
		} catch (Exception e) {
			handleNonFatalException(e);
		}	
	}
	
	
	
	@Override
	public void processResponse(ResponseEvent responseEvent) {
		Response res = responseEvent.getResponse();
		int statusCode = res.getStatusCode();
		
		//Recuperation de la transaction courante
		ClientTransaction transaction = responseEvent.getClientTransaction();
		
		//OK
		if (statusCode == Response.OK) {
			processOK(res, transaction);
			return;
		}

		//MOVED TEMPORARILY
        if (statusCode == Response.BUSY_HERE){
        	processBUSY(res, transaction);
        	return;
        }
		// DECLINE
        if (statusCode == Response.DECLINE) {
        	processDECLINE(res, transaction);
        	return;
        }
		

        if (statusCode >= 400) {
        	chatListener.notify(EventType.ERROR, new String("Error Status" + statusCode));
        }	
        else {
        	chatListener.notify(EventType.ERROR, new String("Bad response type received: " + statusCode));
        }
	}
	

	protected void processOK(Response newOK, ClientTransaction newTransaction) {
		CSeqHeader  cSeqHeader = (CSeqHeader) newOK.getHeader(CSeqHeader.NAME);
		ToHeader toHeader = (ToHeader) newOK.getHeader(ToHeader.NAME);
		Dialog dialog = newTransaction.getDialog();

		// Si le OK provient d'une requte INVITE, confirmation par l'envoi de ACK
		if (cSeqHeader.getMethod().equals(Request.INVITE)) {
			try {
				Request ack = dialog.createAck(cSeqHeader.getSeqNumber());
				dialog.sendAck(ack);
				pendingInvite = false; 
				inviteDialog = dialog;
				otherPeerAddress = toHeader.getAddress();
				String otherPeerURI = otherPeerAddress.getURI().toString();
				chatListener.notify(EventType.SESSION_ESTABLISHED, otherPeerURI);
			}
			catch (Exception ex) {
				handleNonFatalException(ex);
			}
		}		
		//Si le OK provient d'une requte BYE
		else if (cSeqHeader.getMethod().equals(Request.BYE)) {
			// remise de l'agent dans l'tat prt
			inviteDialog = null;
			chatListener.notify(EventType.SESSION_CLOSED, null);
			if (isInSubscribeSession()) {
				sendActiveNotify(true); 
			}
		}
		
		//Si le OK provient d'une requte MESSAGE
		else if (cSeqHeader.getMethod().equals(Request.MESSAGE)) {
			chatListener.notify(EventType.MESSAGE_SENT, null); 
		} 
		// Si le OK provient d'une requte SUBSCRIBE
		else if (cSeqHeader.getMethod().equals(Request.SUBSCRIBE)) {
			// on regarde si c'est une fin de souscription
			ExpiresHeader expiresHeader = newOK.getExpires();
			int expires = expiresHeader.getExpires();
			if (expires > 0) {
				subscribeDialog = dialog;
				chatListener.notify(EventType.INFO, "Subscribe OK");			
			} else {
				subscribeDialog = null;
			}
		} 
		// Si le OK provient d'une requte NOTIFY
		else if (cSeqHeader.getMethod().equals(Request.NOTIFY)) {
			chatListener.notify(EventType.INFO, "Notify");
		}
		// Sinon on ignore
	}

	protected void processBUSY(Response res, ClientTransaction transaction) {
		
		if (!isInSubscribeSession()) {
			AcceptDeclineData data = new AcceptDeclineData(); 
			chatListener.notify(EventType.SESSION_BUSY, data);
		
			// demander  l'utilisateur si il veut faire un souscription automatique
			// se mettre en attente de notify dans ce cas 
			if (data.isAccepted()) {
				sendSubscribe();
			} else {
				chatListener.notify(EventType.SESSION_CLOSED, null);
			}	
		}
	}

	protected void processDECLINE(Response res, ClientTransaction transaction) {
		// on examine si cela concerne la session de conversation ou de  subscription
		if (isChatSessionTransaction(transaction)) {
			chatListener.notify(EventType.SESSION_DECLINED, null);
		}
		else if (isSubscribeSessionTransaction(transaction)) {
			chatListener.notify(EventType.SUBSCRIPTION_DECLINED, null);
		}
	}
	
	
	@Override
    public void processRequest(RequestEvent evt) {
    	Request req = evt.getRequest();
    	String method = req.getMethod();
    	ServerTransaction serverTransaction = evt.getServerTransaction();
		
    	if (serverTransaction == null) {
			try {
				serverTransaction = sipProvider.getNewServerTransaction(req);
			} catch (Exception ex) {
				handleNonFatalException(ex);
			}
		}
    	
    	if (method.equals(Request.ACK)) { 
    		processACK(req, serverTransaction);
    		return;
    	}
    	   	
    	if (method.equals(Request.INVITE)) {
    		processINVITE(req, serverTransaction);
			return; 
    	}

    	if (method.equals(Request.BYE)) {
    		processBYE(req, serverTransaction);
    		return;
    	} 
    		
    	if (method.equals(Request.MESSAGE)) { 
    		processMESSAGE(req, serverTransaction);
    		return;
    	}
    	
    	if (method.equals(Request.SUBSCRIBE)) {
    		processSUBSCRIBE(req, serverTransaction);
    		return;
    	}
    	
    	if (method.equals(Request.NOTIFY)) {
    		processNOTIFY(req, serverTransaction);
    		return;
    	}
    	
    	chatListener.notify(EventType.ERROR, new String("Bad request type received: " + method));
	}

		
	protected void processSUBSCRIBE(Request req, ServerTransaction transaction) {
		
		try {
			Response res;
			// si session de conversation existe
			if (isInChatSession()) {
				// on regarde si on a deja une souscription
				// si ce n'est pas le cas -> on retourne OK avec gestion de temporisation
				if (!isInSubscribeSession()) {
					ExpiresHeader inExpiresHeader = req.getExpires();
					int exp = inExpiresHeader.getExpires();
					res = messageFactory.createResponse(Response.OK, req);
					
					// on place la duree d'expiration dans la rponse et on envoie
					ExpiresHeader  outExpiresHeader = headerFactory.createExpiresHeader(exp);
					res.addHeader(outExpiresHeader);
					transaction.sendResponse(res);
					
					// envoi du notify
					sendActiveNotify(false);
					
					// creation d'une tache pour grer l'expiration
					Timer timer = new Timer(); 
					timerTask = new TimerTask() { 
							@Override
							public void run() {
								sendTerminatedNotify();
								subscribeDialog = null;
							}
					};
					timer.schedule(timerTask, exp * 1000);
					chatListener.notify(EventType.SUBSCRIPTION_ACCEPTED, null);
				} else {
					// on regarde si le message SUBSCRIBE concerne la session en cours
					if (isSubscribeSessionTransaction(transaction)) {
						ExpiresHeader inExpiresHeader = req.getExpires();
						int exp = inExpiresHeader.getExpires();
						
						// soit un arret de souscription, soit une prolongation
						// on arrete le timer dans les deux cas
						timerTask.cancel();
						
						if (exp > 0) {
							// creation d'une tache pour grer la nouvelle dure d'expiration
							Timer timer = new Timer(); 
							timerTask = new TimerTask() { 
									@Override
									public void run() {
										sendTerminatedNotify();
										subscribeDialog = null;
									}
							};
							timer.schedule(timerTask, exp * 1000);
						}
						else {
							chatListener.notify(EventType.SUBSCRIPTION_CLOSED, null);
						}
						res = messageFactory.createResponse(Response.OK, req);
						transaction.sendResponse(res);
					}	
					else {
						// sinon on decline la souscription
						res = messageFactory.createResponse(Response.DECLINE, req);

					}
				}
			}
			else {
				// sinon pas de session chat c'est une mauvaise requete
				res = messageFactory.createResponse(Response.BAD_REQUEST, req);
				transaction.sendResponse(res);			
			}
			
		} catch(Exception e) {
			handleNonFatalException(e);
		} 
	}
	
	

	protected void processNOTIFY(Request req, ServerTransaction transaction) {
		try {
			// on teste si une session de conversation existe
			if (!isInChatSession()) {
				// pas de session de chat mais peut tre une session de souscription 
				if (isInSubscribeSession() && isSubscribeSessionTransaction(transaction)) {
					Response res = messageFactory.createResponse(Response.OK, req);
					transaction.sendResponse(res);
					
					String content = new String(req.getRawContent());
					if (content.equals("available")) {
						sendTerminateSubscribe(); 
						sendInvite(otherPeerAddress.getURI().toString());
					}
					return;
				}
			} 
			// mauvaise requte car pas de session de souscription ou 
			// requte hors session de souscription ou session de conversation existante
			Response res = messageFactory.createResponse(Response.BAD_REQUEST, req);
			transaction.sendResponse(res);
		}
		catch(Exception e) {
			handleNonFatalException(e);
		}
	}
	
	
    public void processMESSAGE(Request newMessage, ServerTransaction transaction) {
    	try {
    		Response response = null;
        	
        	// verification que le message concerne la session en cours
        	if (isChatSessionTransaction(transaction)) {      		
        		response = messageFactory.createResponse(Response.OK, newMessage);
        		transaction.sendResponse(response);
        		chatListener.notify(EventType.MESSAGE_RECEIVED, new String(newMessage.getRawContent()));
        	} else {
        		// sinon on indique un mauvaise requte	
    			response = messageFactory.createResponse(Response.BAD_REQUEST, newMessage);		
    			transaction.sendResponse(response);
        	}
    		
    	} catch(Exception e) {
    		handleNonFatalException(e);
    	}
    }
        

    protected void processINVITE(Request newInvite, ServerTransaction transaction) {
		try {
			CallIdHeader callIDHeader = (CallIdHeader) newInvite.getHeader(CallIdHeader.NAME);
			CSeqHeader cSeqHeader = (CSeqHeader) newInvite.getHeader(CSeqHeader.NAME);
			FromHeader fromHeader = (FromHeader) newInvite.getHeader(FromHeader.NAME);
			Response res; 
			
			// on verifie la destination sinon on repond DOES_NOT_EXIST_ANYWHERE
			ToHeader to = (ToHeader) newInvite.getHeader("To");
        	if (! to.getAddress().getURI().equals(myAddress.getURI())) {
        		res = messageFactory.createResponse(Response.DOES_NOT_EXIST_ANYWHERE, newInvite);
        		transaction.sendResponse(res);
        		return;
        	}
      			
			if (!isInChatSession()) {
				if (pendingInvite) {
					// si en attente de confirmation, on reponse DECLINE
					res = messageFactory.createResponse(Response.DECLINE, newInvite);
					transaction.sendResponse(res);
				}
				else {
					// si pas d'invite en cours
					String sender = fromHeader.getAddress().getURI().toString();
					AcceptDeclineData data = new AcceptDeclineData(sender);
					chatListener.notify(EventType.SESSION_REQUESTED, data); 
					
				
					if (data.isAccepted()) {
						// on renvoie un OK si utilisateur accepte l'invitation 
						res = messageFactory.createResponse(Response.OK, newInvite);
						ToHeader toHeader = (ToHeader) res.getHeader(ToHeader.NAME);
						//gnration du LocalTag serveur (diffrent du client)
						String tag = genTag();
						toHeader.setTag(tag);
						pendingInvite = true;
						
						ContactHeader contactHeader = headerFactory.createContactHeader(myAddress);
						res.addHeader(contactHeader);
						
						inviteDialog = transaction.getDialog();
						
						// si une souscription en cours, on donne la priorit  l'invitation et on annule
						if (isInSubscribeSession()) {
							sendTerminateSubscribe();
						}
 						
					} else {
						res = messageFactory.createResponse(Response.DECLINE, newInvite);
					}
					transaction.sendResponse(res);
				} 
			}
			else {
				// on est dj dans une session et on teste si c'est une une nouvelle 
				if (!isChatSessionTransaction(transaction))  {
					// nouvel session
					if (!isInSubscribeSession()) {
						res = messageFactory.createResponse(Response.BUSY_HERE, newInvite);
						transaction.sendResponse(res);
						// message processor 
						chatListener.notify(EventType.SESSION_REQUESTED_ONBUSY, fromHeader.getAddress().getURI().toString());
					}
					else {
						res = messageFactory.createResponse(Response.DECLINE, newInvite);
						transaction.sendResponse(res);
					}
				}
				// sinon c'est un doublon, on ignore
			}
		}
		catch (Exception e) {
			handleReInitException(e);
		}
	}

    /**
	 * Genere un nouveau tag
	 * @result un tag genere aleatoirement
	 */
	protected String genTag() {
		String tag = Integer.toHexString(new Random().nextInt());
		return tag;
	}

    
	protected void processACK(Request newAck, ServerTransaction transaction) {
    	
		if (transaction != null) {
			inviteDialog = transaction.getDialog();
			pendingInvite = false; 
			
			FromHeader fromHeader = (FromHeader) newAck.getHeader(FromHeader.NAME);
			otherPeerAddress = fromHeader.getAddress();
			String otherPeerURI = otherPeerAddress.getURI().toString();
			chatListener.notify(EventType.SESSION_ESTABLISHED, otherPeerURI);
    	} else {
    		chatListener.notify(EventType.ERROR, new String("Incorrect ACK message"));
    	}
	}
    

	protected void processBYE(Request newBye, ServerTransaction transaction) {

		try {
			Response res = null; 
			// on verifie que le BYE concerne la session courante
			if (isChatSessionTransaction(transaction)) {
				// on confirme la fin de session avec un OK
				res = messageFactory.createResponse(Response.OK, newBye);
				transaction.sendResponse(res);
				chatListener.notify(EventType.SESSION_CLOSED, null); 
			
				// si il y a une souscription, on envoie un NOTIFY[available]
				// pour pouvoir tre invite par l'utilisateur ayant souscrit
				if (subscribeDialog != null) {
					sendActiveNotify(true);
				}
			
				// on remet dans l'tat prt
				inviteDialog = null;
			} else {
				// sinon erreur
				res = messageFactory.createResponse(Response.BAD_REQUEST, newBye);
				transaction.sendResponse(res);
			}
		} catch (Exception e) {
			handleReInitException(e);
		}

	}
    	

	@Override
    public void processTimeout(TimeoutEvent evt) {
    	chatListener.notify(EventType.ERROR, new String("Timeout"));
    }

    @Override
    public void processIOException(IOExceptionEvent evt) {
    	chatListener.notify(EventType.ERROR, new String("I/O Exception"));
    }

    @Override
    public void processDialogTerminated(DialogTerminatedEvent evt) {
    	System.out.println("Dialog terminated");
    }

    @Override
    public void processTransactionTerminated(TransactionTerminatedEvent evt) {
    }

    public String getHost() {
    	String host = sipStack.getIPAddress();
    	return host;
    }

    public int getPort() {
    	int port = listeningPoint.getPort();
    	return port;
    }

    public String getUsername() {
    	return username;
    }

    public void setUsername(String newUsername) {
	username = newUsername;
    }

    public ChatSipEventListener getChatListener() {
	return chatListener;
    }

    public void setChatListener(ChatSipEventListener listener) {
    	chatListener = listener;
    }

	public boolean isConnected() {
		return inviteDialog != null;
	}
	
	protected void handleFatalException(Exception e) {
		e.printStackTrace();
		System.exit(0);
	}
	protected void handleReInitException(Exception e) {
		e.printStackTrace();
		
		// reinitialisation de l'tat
		if (timerTask != null) {
			timerTask.cancel();
			timerTask = null;
		}
		if (subscribeDialog != null) {
			subscribeDialog.delete();
			subscribeDialog = null;
		}
		if (inviteDialog != null) {
			inviteDialog.delete();
			inviteDialog = null;
		}
		pendingInvite = false;
		
 	}
	
	protected void handleNonFatalException(Exception e) {
		e.printStackTrace(); 
	}

}
