package livresip.proxy;

import gov.nist.javax.sip.header.SIPHeader;

import java.net.InetAddress;
import java.text.ParseException;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;
import java.util.ListIterator;
import java.util.Properties;
import java.util.Random;
import java.util.Timer;
import java.util.TimerTask;
import java.util.TooManyListenersException;
import java.util.Vector;

import javax.mail.Authenticator;
import javax.mail.MessagingException;
import javax.mail.PasswordAuthentication;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;
import javax.sip.ClientTransaction;
import javax.sip.Dialog;
import javax.sip.DialogTerminatedEvent;
import javax.sip.IOExceptionEvent;
import javax.sip.InvalidArgumentException;
import javax.sip.ListeningPoint;
import javax.sip.ObjectInUseException;
import javax.sip.PeerUnavailableException;
import javax.sip.RequestEvent;
import javax.sip.ResponseEvent;
import javax.sip.ServerTransaction;
import javax.sip.SipFactory;
import javax.sip.SipListener;
import javax.sip.SipProvider;
import javax.sip.SipStack;
import javax.sip.Transaction;
import javax.sip.TransactionTerminatedEvent;
import javax.sip.TransportNotSupportedException;
import javax.sip.address.Address;
import javax.sip.address.AddressFactory;
import javax.sip.address.SipURI;
import javax.sip.address.URI;
import javax.sip.header.CSeqHeader;
import javax.sip.header.ContactHeader;
import javax.sip.header.FromHeader;
import javax.sip.header.HeaderFactory;
import javax.sip.header.MaxForwardsHeader;
import javax.sip.header.RecordRouteHeader;
import javax.sip.header.RouteHeader;
import javax.sip.header.ToHeader;
import javax.sip.header.ViaHeader;
import javax.sip.message.Message;
import javax.sip.message.MessageFactory;
import javax.sip.message.Request;
import javax.sip.message.Response;

import org.apache.log4j.BasicConfigurator;
import org.apache.log4j.Level;
import org.apache.log4j.Logger;

/**
 * 	Simple Sip Proxy 
 * 
 *  @see Chapitre 7 du livre Tlphonie SIP : concepts, usages et programmation en Java
 *	@author G. Vanwormhoudt
 */
 
public class ProxyServer implements SipListener {

	private static final int VALID_CODE = 0;
	protected SipFactory sipFactory;
	protected AddressFactory addressFactory;
	protected MessageFactory messageFactory;
	protected HeaderFactory headerFactory;
	protected SipStack sipStack;

	protected static Logger logger;
	
	protected SipProvider sipProvider;
	protected ListeningPoint listeningPoint;
	
	protected int port;
	protected String transport;

	private static Random localTagGenerator = new Random();
	
	protected HashMap<URI, ContactHeader> registrations;
	
	protected HashMap<URI, Timer> timers; 

	
	public ProxyServer(String ip, int port, String transport) throws TransportNotSupportedException, 
		InvalidArgumentException, ObjectInUseException, TooManyListenersException, PeerUnavailableException {
			
		initSip(ip, port, transport);
		initLogger(); 
		
		registrations = new HashMap<URI, ContactHeader>();
		timers = new HashMap<URI, Timer>();
	}

	private void initSip(String ip, int port, String transport)
			throws PeerUnavailableException, TransportNotSupportedException,
			InvalidArgumentException, ObjectInUseException,
			TooManyListenersException {
		sipFactory = SipFactory.getInstance();
		sipFactory.setPathName("gov.nist");
		Properties properties = new Properties();
		properties.setProperty("javax.sip.STACK_NAME", "Proxy");
		properties.setProperty("javax.sip.IP_ADDRESS", ip);
		// properties.setProperty("javax.sip.AUTOMATIC_DIALOG_SUPPORT", "off");

		sipStack = sipFactory.createSipStack(properties);
			
		headerFactory = sipFactory.createHeaderFactory();
		addressFactory = sipFactory.createAddressFactory();
		messageFactory = sipFactory.createMessageFactory();
		
		this.port = port;
		this.transport = transport;
	}
	
	private void initLogger() {
		BasicConfigurator.configure();	
		logger = Logger.getLogger(ProxyServer.class.getName());
		logger.setLevel(Level.INFO);
	}


	public void processRequest(RequestEvent requestEvent) {
		Request request = requestEvent.getRequest();
		ServerTransaction serverTransaction = requestEvent.getServerTransaction();
		String method = request.getMethod();
			
		logger.info("Request " + method + " received with server transaction id " + serverTransaction);
		try {
			if (method.equals(Request.REGISTER)) {
				processRegister(requestEvent, serverTransaction);
			} 
			else {
				processOtherRequests(requestEvent, serverTransaction);
			}
		} catch(Exception ex) {
			logger.error("Internal Error raised", ex);
			replyInternalError(request, serverTransaction);
		} 
	}

	protected void replyError(Request request,
			ServerTransaction serverTransaction, int errorCode) {
		try {
			Response response = messageFactory.createResponse(errorCode, request);
			if (serverTransaction != null) {
				serverTransaction.sendResponse(response);
			} else {
				sipProvider.sendResponse(response);
			}
		} catch(Exception ex) {
			ex.printStackTrace();
		}
	}
	
	protected void replyInternalError(Request request, ServerTransaction serverTransaction) {
		replyError(request, serverTransaction, Response.SERVER_INTERNAL_ERROR);
	}

	
	public void processOtherRequests(RequestEvent requestEvent, ServerTransaction serverTransaction) {

		Request request = requestEvent.getRequest();
		SipProvider sipProvider = (SipProvider) requestEvent.getSource();
		String method = request.getMethod();
		
		try {

			// Vrification de la requte
			int errorCode = validateRequest(request);
			if (errorCode != VALID_CODE) {
				replyError(request, serverTransaction, errorCode);
				logger.info("processRequest - Request not validated ");
				return;
			}

			// Si c'est un ACK  destination du proxy, on ne fait plus rien
			if (method.equals(Request.ACK)) {
				ListIterator routes = request.getHeaders(RouteHeader.NAME);
				if (routes == null || !routes.hasNext()) {
					logger.info("processRequest ACK targeted for the proxy, Ignore it");
					return;
				}
			}

			// Suppression de l'adresse du proxy dans le champ ROUTE si elle est presente
			ListIterator routes = request.getHeaders(RouteHeader.NAME);
			if (routes != null) {
				if (routes.hasNext()) {
					RouteHeader routeHeader = (RouteHeader) routes.next();
					Address routeAddress = routeHeader.getAddress();
					SipURI routeSipURI = (SipURI) routeAddress.getURI();

					String host = routeSipURI.getHost();
					int port = routeSipURI.getPort();
					if (listeningPoint.getPort() == port
						&& listeningPoint.getIPAddress().equals(host)) {
							routes.remove();
							logger.info("Remove the first route from the RouteHeader: matches the proxy:"+ routes);
						}
					}
			}  

			if (serverTransaction == null && !method.equals(Request.ACK)) {
				serverTransaction = sipProvider.getNewServerTransaction(request);
			}		
			
			URI requestURI = request.getRequestURI();
		
			ContactHeader contactHeader = registrations.get(requestURI);
			URI targetURI = contactHeader != null ? contactHeader.getAddress().getURI() : null;
			 
			// Si adresse logique inconnu, on renvoie DOES_NOT_EXIST_ANYWHERE si emetteur inconnu
			if (targetURI == null) {
				FromHeader from = (FromHeader) request.getHeader(SIPHeader.FROM);
				URI fromURI = from.getAddress().getURI();
				ContactHeader fromContact = registrations.get(fromURI);
				if (fromContact == null) {
					Response response = messageFactory.createResponse(Response.DOES_NOT_EXIST_ANYWHERE , request);
					serverTransaction.sendResponse(response);
					logger.info("Sending a DOESNOTEXIST response"); 
					return;
				} else {
					targetURI = requestURI;
					logger.info("Keep target:" + targetURI);
				}
			} else {
				logger.info("Found target:" + targetURI);			
			}	
				
			// Si BYE, on revoit toute de suite le OK au client avant la redirection		
			if (method.equals(Request.BYE)) {
				if (serverTransaction == null) {
					logger.info("processRequest null server transactoin for BYE");
					return;
				}
				Response response = messageFactory.createResponse(
						Response.OK, request);
				logger.info("Out-Response :" + response);
				serverTransaction.sendResponse(response);
			}
			
			// Si INVITE repond TRYING et redirige le INVITE
			if (method.equals(Request.INVITE)) {
				Response response = messageFactory.createResponse(Response.TRYING, request);
				serverTransaction.sendResponse(response);
				logger.info("sending a TRYING response");
			}
			
			// Si CANCEL, traitement en Hop by Hop
			if (method.equals(Request.CANCEL)) {
				processCancelHopByHop(request, serverTransaction, targetURI);
				return;
			}
				
			this.forwardRequest(request, serverTransaction, targetURI);

		} catch (Exception ex) {
			logger.error("Internal Error raised", ex);
			replyInternalError(request, serverTransaction);
		}
	}


	protected void processCancelHopByHop(Request request,
			ServerTransaction serverTransaction, URI targetURI) {
		
		try {
			Response response;
			
			// En premier, on envoie un rponse OK pour le CANCEL 
			response = messageFactory.createResponse(Response.OK, request);	
			serverTransaction.sendResponse(response);
			logger.info("200 OK response to CANCEL: " + response);
			
			// Suivi d'une rponse 487 Request Terminated to reply pour le INVITE
			Dialog serverDialog = serverTransaction.getDialog();
			// Rcupration de la transaction range dans l'objet Dialog 
			ServerTransaction inviteTransaction = (ServerTransaction) serverDialog.getApplicationData();

			Request inviteRequest = inviteTransaction.getRequest();
			response =  messageFactory.createResponse(Response.REQUEST_TERMINATED, inviteRequest); 			
			inviteTransaction.sendResponse(response);
			logger.info("487 Request Terminated  response to INVITE: " + response);
					
			// Envoi du CANCEL au destinataire 
			ClientTransaction clientTransaction = ((ServerTransactionData) inviteTransaction.getApplicationData()).clientTransaction;		
			// Rcupration de la transaction associe  la propagation du INVITE
			Request cancel = clientTransaction.createCancel();
			ClientTransaction newClientTransaction = sipProvider.getNewClientTransaction(cancel);
			newClientTransaction.sendRequest();
			logger.info("Sending request:" + cancel + " from client transaction:" + clientTransaction);
				
		
		} catch (Exception ex) {
			logger.error("Internal Error raised", ex);
			replyInternalError(request, serverTransaction);
		}
	}

	public static synchronized String generateLocalTag() {
		return Integer.toHexString(localTagGenerator.nextInt());
	}
	

	public int validateRequest(Request request) {
		int resultCode = VALID_CODE; 
		
		if (!checkURIScheme(request)) {
			resultCode = Response.UNSUPPORTED_URI_SCHEME;
		}
		else {
			if (!checkMaxForwards(request)) {
			resultCode = Response.TOO_MANY_HOPS; 
			}
			else {
				if (!checkLoopDetection(request)) {
					resultCode = Response.LOOP_DETECTED;
				}
				else {
					if (!checkFromTag(request)) {
						resultCode = Response.BAD_REQUEST;
					}
				}
			}
		}
		return resultCode;
	}
	

	protected boolean checkURIScheme(Request request) {
		URI requestURI = request.getRequestURI();
		String uriScheme = requestURI.getScheme();
		return uriScheme.equals("sip") || uriScheme.equals("sips") || uriScheme.equals("tel");
	}

	protected boolean checkMaxForwards(Request request) {
		MaxForwardsHeader mf = (MaxForwardsHeader) request.getHeader(MaxForwardsHeader.NAME);
		if (mf == null) {
			return true;
		}
		if (mf.getMaxForwards() == 0) {
			return false;
		} else {
			return true;
		}
	}

	protected boolean checkLoopDetection(Request request) {
		ListIterator viaList = request.getHeaders(ViaHeader.NAME);
		if (viaList == null)
			return false;

		String stackIPAddress = sipStack.getIPAddress();

		while (viaList.hasNext()) {
			ViaHeader viaHeader = (ViaHeader) viaList.next();
			String host = viaHeader.getHost();
			if (host.equals(stackIPAddress)) {
				Iterator lps = sipStack.getListeningPoints();
				while (lps != null && lps.hasNext()) {
					ListeningPoint lp = (ListeningPoint) lps.next();
					int port = lp.getPort();
					if (viaHeader.getPort() == port) {
						logger.info("RequestValidation, checkLoopDetection(),"
								+ " the via port matches the proxy port");
						return false;
					}
				}
			}
		}
		return true;
	}

	protected boolean checkFromTag(Request request) {
		if (request.getMethod().equals("REGISTER"))
			return true;
		FromHeader fh = (FromHeader) request.getHeader(FromHeader.NAME);
		return (fh.getTag() != null);
	}
	
		
	protected void processRegister(
		RequestEvent requestEvent,
		ServerTransaction serverTransaction) {
		
		Request request = requestEvent.getRequest();
		try {			
			ToHeader toHeader = (ToHeader) request.getHeader(ToHeader.NAME);
			ContactHeader contactHeader = (ContactHeader) request.getHeader(ContactHeader.NAME);
				
			int expires = contactHeader.getExpires();
			
			// store To and Contact information in registration table 
			URI toURI = toHeader.getAddress().getURI();
			
			ContactHeader registeredContact = registrations.get(toURI);
			
			if (registeredContact == null) {
				// new entry
				if (expires != 0) {
					registrations.put(toURI, contactHeader);
					createRegisterTimer(toURI, expires);
					logger.info("New registration:" + toHeader.getAddress() + "->" + contactHeader.getAddress());
				} 
			} 
			else {
				// existing entry
				registrations.remove(toURI);
				deleteRegisterTimer(toURI);
				if (expires != 0) {
					registrations.put(toURI, contactHeader);
					createRegisterTimer(toURI, expires);
				} else {
					logger.info("Remove registration:" + toURI);
				}
			}
			
			// sending ok with expire and tag insertion
			Response response = messageFactory.createResponse(Response.OK, request);
			// clone the contact header 
			ContactHeader newContactHeader = (ContactHeader) contactHeader.clone();
			newContactHeader.setExpires(3600); 
			response.setHeader(contactHeader);
			
			toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
			toHeader.setTag(generateLocalTag());
						
			ServerTransaction newServerTransaction = sipProvider.getNewServerTransaction(request);
			newServerTransaction.sendResponse(response);
			
		} catch (Exception ex) {
			logger.error("Internal Error raised", ex);
			replyInternalError(request, serverTransaction);

		}
	}
	
	protected void createRegisterTimer(URI uri, int expires) {
		Timer timer = new Timer(); 
		RegisterTimerTask task = new RegisterTimerTask(uri);
		int delay = expires * 1000;
		timer.schedule(task, delay);
		timers.put(uri, timer);
	}
	
	protected void deleteRegisterTimer(URI uri) {
		Timer timer = timers.get(uri);
		timer.cancel(); 
		timers.remove(uri);
	}

	public void processResponse(ResponseEvent responseEvent) {
			
		Response response = (Response) responseEvent.getResponse();
		ClientTransaction clientTransaction = responseEvent.getClientTransaction();
		int statusCode = response.getStatusCode();

		logger.info("Response " + response.getStatusCode() + " received with client transaction id " + clientTransaction);
		try {
			 if (statusCode == Response.RINGING) {
				processRinging(response, clientTransaction); 
			} else if (statusCode == Response.BUSY_HERE) {
				processBusyHere(response, clientTransaction);
			} else if ((statusCode == Response.REQUEST_TERMINATED) || (statusCode == Response.TRYING) ||
					   (statusCode == Response.OK &&  clientTransaction.getRequest().getMethod().equals(Request.CANCEL))) {
				return;
			} else {
				forwardResponse(response, clientTransaction);
			}	
		} 
		catch(Exception ex) {
			logger.error("Internal Error raised", ex);
		} 	
	}
		
	
	public void processBusyHere(Response response, ClientTransaction clientTransaction) {
		try {
			if (clientTransaction != null) {
				// retransmission du BUSY 
				forwardResponse((Response) response.clone(), clientTransaction); 
				sendCallMail(response);
			}			
			else {
				forwardResponse((Response) response.clone(), clientTransaction);		
			}							
		}
		catch (Exception ex) {
			logger.error("Internal Error raised", ex);
		}
	}
	
	protected void processRinging(Response response, ClientTransaction clientTransaction) {			
		try {
			if (clientTransaction != null) {
				// retransmission du Ringing 
				CSeqHeader cseq = (CSeqHeader) response.getHeader(CSeqHeader.NAME);
				if (cseq.getMethod().equals(Request.INVITE)) {
					ClientTransactionData clientTransactionData = (ClientTransactionData) clientTransaction.getApplicationData();
					if (clientTransactionData  != null)  {
						clientTransactionData.ringingCount--;
						// on teste si on atteint le nombre de ringing pour prendre une dcision
						if (clientTransactionData.ringingCount != 0) {
							// retransmission du Ringing
							forwardResponse((Response) response.clone(), clientTransaction); 
						}
						else {
							// on retourn une rponse Temporarily Unavailable au client
							Response newResponse = (Response) response.clone();
							newResponse.setStatusCode(Response.TEMPORARILY_UNAVAILABLE);
							forwardResponse(newResponse, clientTransaction); 
							
							// on envoie un cancel dans ce cas pour stopper l'tablissement session
							Request cancelRequest = clientTransaction.createCancel();
							ClientTransaction cancelTransaction = sipProvider.getNewClientTransaction(cancelRequest);
						    cancelTransaction.sendRequest();
						}
					}
				}
			} 
			else {
				forwardResponse((Response) response.clone(), clientTransaction);
			}
		}
		catch (Exception ex) {
			logger.error("Internal Error raised", ex);
		}
		
	}

	public void forwardResponse(Response response,ClientTransaction clientTransaction) {
	
		logger.info("Forwarding response:" + response + " with client transaction:" + clientTransaction);

		CSeqHeader cseqHeader = (CSeqHeader) response
				.getHeader(CSeqHeader.NAME);
		String method = cseqHeader.getMethod();
		int statusCode = response.getStatusCode();
		String reason = response.getReasonPhrase();
		
		boolean hasTopViaProxy = hasTopViaHeaderProxy(response);

		try {
			ListIterator viaList = response.getHeaders(ViaHeader.NAME);
			// Suppression du premier champ Via correspondant au proxy
			if (hasTopViaProxy) {
				response.removeHeader(ViaHeader.NAME);
				Vector v = new Vector();
				viaList.next();
				while (viaList != null && viaList.hasNext()) {
					ViaHeader viaHeader = (ViaHeader) viaList.next();
					v.addElement(viaHeader);
				}
				for (int j = 0; j < v.size(); j++) {
					ViaHeader viaHeader = (ViaHeader) v.elementAt(j);
					response.addHeader(viaHeader);
				}		
			}
			
			viaList = response.getHeaders(ViaHeader.NAME);
			// Si il reste au moins un champ via aprs la suppression
			if (viaList != null && viaList.hasNext()) {
				if (clientTransaction == null) {
 					sipProvider.sendResponse(response);
				}
				else {
					// Propagation de la reponse en recuperant la transaction serveur
					ClientTransactionData clientTransactionData = (ClientTransactionData) clientTransaction.getApplicationData();
					ServerTransaction serverTransaction = null; 
					if (clientTransactionData != null) {
						serverTransaction = clientTransactionData.serverTransaction;	
						if (serverTransaction == null) {
							sipProvider.sendResponse(response);
						} else {
							serverTransaction.sendResponse(response);
						}
					}
				}
			} 
			else {
				// on abandonne la rponse
			}
		} catch (Exception ex) {
			logger.error("Internal error during reponse forwarding",  ex);
		}
	}
	
	
	public boolean hasTopViaHeaderProxy(Message message) {
		ListIterator viaList = message.getHeaders(ViaHeader.NAME);
		if (viaList == null)
			return false;

		String stackIPAddress = sipStack.getIPAddress();
		ViaHeader viaHeader = (ViaHeader) viaList.next();
		
		if (viaHeader.getHost().equals(stackIPAddress)) {
			Iterator lps = sipStack.getListeningPoints();
			while (lps != null && lps.hasNext()) {
				ListeningPoint lp = (ListeningPoint) lps.next();
				int port = lp.getPort();
				if (viaHeader.getPort() == port)
					return true;
			}
		}
		return false;
	}
	
	
	
	
	protected void sendCallMail(Response response) {
		// envoi du mail avec javax.mail 
		// rcupration de l'emetteur  partir du champ Contact 
		FromHeader fromHeader = (FromHeader) response.getHeader(FromHeader.NAME); 
		ToHeader toHeader = (ToHeader) response.getHeader(ToHeader.NAME);
		URI toURI = toHeader.getAddress().getURI();
		
		ContactHeader contactHeader = registrations.get(toURI);
		String mail = null; 
		if (contactHeader != null) {
			mail = contactHeader.getParameter("usermail");
			mail = mail + "@" + contactHeader.getParameter("domainmail");
		}
		if (mail != null) {
			Properties props = new Properties();
			props.put("mail.smtp.host", "smtp.telecom-lille1.eu");
			props.put("mail.smtp.auth", "true");

		    Authenticator auth = new SMTPAuthenticator();
		    Session session = Session.getDefaultInstance(props, auth);
	
			try {
				MimeMessage msg = new MimeMessage(session);
				msg.setFrom();
				msg.setFrom(new InternetAddress("proxyserv@telecom-lille1.eu"));
			    msg.setRecipients(MimeMessage.RecipientType.TO, mail);
			    msg.setSubject("Appel manqu le "+ new Date());
			    msg.setSentDate(new Date());
			    msg.setText("Vous avez reu un appel de \n" + fromHeader.getAddress());
			    Transport.send(msg);
			} catch (MessagingException mex) {
			    logger.info("send failed, exception: " + mex);
			}
		} else {
			logger.info("No adress found for sending email");
		}
	}
		
	

	protected void forwardRequest(Request request, ServerTransaction serverTransaction, URI targetURI) {			
		try {
			
			Request preparedRequest = prepareForward(request, targetURI);
			
			logger.info("Forwarding request:" + preparedRequest + " from server transaction:" + serverTransaction);
			
			String method = preparedRequest.getMethod();
			
			// Si c'est un ACK, on propage sans tat
			if (method.equals(Request.ACK)) {
				sipProvider.sendRequest(preparedRequest);
				return;
			}	
	
			ClientTransaction clientTransaction = sipProvider.getNewClientTransaction(preparedRequest);
			clientTransaction.sendRequest();
			
			// enregistrement de la liaison entre les deux transactions
			serverTransaction.setApplicationData(new ServerTransactionData(clientTransaction));
			clientTransaction.setApplicationData(new ClientTransactionData(serverTransaction));
			
			// enregistrement de la liaison entre la transaction serveur et le dialogue
			if (method.equals(Request.INVITE)) {
				Dialog serverDialog = sipProvider.getNewDialog(serverTransaction);
				serverDialog.setApplicationData(serverTransaction);
			}
			
		} catch (Exception ex) {
			logger.error("Internal Error raised", ex);
			replyInternalError(request, serverTransaction);
		} 	
	}

	
	
		
	protected Request prepareForward(Request request, URI targetURI) {	
		Request clonedRequest = (Request) request.clone();
		
		try {
			Iterator parametersNames = null;
			
			URI requestURI = request.getRequestURI();
			// Mise  jour de l'URI de la requtr
			if (! targetURI.equals(requestURI)) {
				clonedRequest.setRequestURI(targetURI);
			}
			prepareMaxForwards(clonedRequest);	
			
			ListeningPoint lp = sipProvider.getListeningPoint("udp");
			String stackIP = lp.getIPAddress();
			prepareVia(clonedRequest, stackIP, lp);
			prepareRouting(clonedRequest, stackIP, lp);
			
		} catch (Exception ex) {
			logger.error("Internal Error raised", ex);
		}
		return clonedRequest;
	}

	protected void prepareMaxForwards(Request clonedRequest)
			throws InvalidArgumentException {
		// Decrementation du champ Max-Forwards
		MaxForwardsHeader mf = (MaxForwardsHeader) clonedRequest.getHeader(MaxForwardsHeader.NAME);
		if (mf == null) {
			mf = headerFactory.createMaxForwardsHeader(70);
			clonedRequest.addHeader(mf);
		} else {
			mf.setMaxForwards(mf.getMaxForwards() - 1);
		}
	}

	protected void prepareVia(Request clonedRequest, String stackIP, ListeningPoint lp) throws ParseException,
			InvalidArgumentException {
		// Ajout du champ Via
		ViaHeader viaHeader = null;
		
		// Traitement particulier en cas de Cancel  cause du mode Hop-By-Hop
		if (clonedRequest.getMethod().equals(Request.CANCEL)) {
			viaHeader = headerFactory.createViaHeader(stackIP,lp.getPort(), lp.getTransport(), null);
			// On ne garde que le champ Via cr
			clonedRequest.removeHeader(ViaHeader.NAME);
		} else {
			viaHeader = headerFactory.createViaHeader(stackIP, lp.getPort(), lp.getTransport(), null);
		}

		// Ajout du champ Via
		if (viaHeader != null) {
			clonedRequest.addHeader(viaHeader);
		}
	}

	protected void prepareRouting(Request request, String stackIP, ListeningPoint lp)
	throws ParseException {

		// Traitement du champ Route
		// Suppression du champ Route si il correspond au proxy
		ListIterator routes = request.getHeaders(RouteHeader.NAME);
		if (routes != null && routes.hasNext()) {
			RouteHeader routeHeader = (RouteHeader) routes.next();
			Address routeAddress = routeHeader.getAddress();
			URI routeURI = routeAddress.getURI();
		
			if (routeURI.isSipURI()) {				
				SipURI routeSipURI = (SipURI) routeURI;

				String host = routeSipURI.getHost();
				int port = routeSipURI.getPort();
		
				if (stackIP.equals(host) && lp.getPort() == port) {
					routes.remove();
				}
			} 
		}				
		
		// Mise  jour du champ Record-route et ajout si c'est un Invite
		if (request.getMethod().equals(Request.INVITE)) {
			SipURI sipURI = addressFactory.createSipURI(null,stackIP);
			sipURI.setPort(lp.getPort());
			sipURI.setLrParam();
			
			Address address = addressFactory.createAddress(null, sipURI);
			RecordRouteHeader recordRouteHeader = headerFactory.createRecordRouteHeader(address);
	
			// On procde par suppression puis ajout des champs Record-Route 
			ListIterator recordRouteHeaders = request.getHeaders(RecordRouteHeader.NAME);
			request.removeHeader(RecordRouteHeader.NAME);   
			Vector v = new Vector();
			v.addElement(recordRouteHeader);
			
			while (recordRouteHeaders != null && recordRouteHeaders.hasNext()) {
				recordRouteHeader = (RecordRouteHeader) recordRouteHeaders.next();
				v.addElement(recordRouteHeader);
			}
			
			for (int j = 0; j < v.size(); j++) {
				recordRouteHeader = (RecordRouteHeader) v.elementAt(j);
				request.addHeader(recordRouteHeader);
			}
		}
	}
	

	public void processTimeout(javax.sip.TimeoutEvent timeoutEvent) {
		Transaction transaction;
		if (timeoutEvent.isServerTransaction()) {
			transaction = timeoutEvent.getServerTransaction();
		} else {
			transaction = timeoutEvent.getClientTransaction();
		}
		logger.info("Transaction Time out for " + transaction + 
				"with state " + transaction.getState());
	}

	
	public void processIOException(IOExceptionEvent arg0) {
		logger.error("Unexpected IO Exception!!");

	}

	public void processTransactionTerminated(TransactionTerminatedEvent txe) {
		logger.info("transaction terminated event");

	}

	public void processDialogTerminated(DialogTerminatedEvent event) {
		logger.info("Dialog terminated event for: " + event.getDialog());

	}
	
	
	public void start() {
		try {
			listeningPoint = sipStack.createListeningPoint(sipStack
					.getIPAddress(), this.port, this.transport);
			sipProvider = sipStack.createSipProvider(listeningPoint);
			sipProvider.setAutomaticDialogSupportEnabled(false);
			sipProvider.addSipListener(this);
			sipStack.start();
			logger.info("Proxy started...");
			logger.info("Now listening at :" + sipStack.getIPAddress() + ":"
					+ listeningPoint.getPort() + "/"
					+ listeningPoint.getTransport());
		} catch (Exception e) {
			logger.error("Exception raised:" + e);
		}
	}
	
	public void stop() throws Exception {
		if (sipStack == null)
			return;
		sipProvider.removeSipListener(this);
		sipStack.deleteSipProvider(sipProvider);
		sipStack.deleteListeningPoint(listeningPoint);
		sipStack.stop();
		logger.info("Proxy stop...");
	}
	
	
	public static void main(String args[]) throws Exception {
		if (args.length != 1) {
			System.err.println("ce programme a besoin d'un numro de port");
		} else {
			String localIP = InetAddress.getLocalHost().getHostAddress();
			int port = Integer.parseInt(args[0]);
			ProxyServer proxy = new ProxyServer(localIP, port, "udp");	
			proxy.start();
		}
	}

	
	/**
	 * Classes pour corrler les transactions  
	 * 
	 */
	
	class ServerTransactionData { 
		ClientTransaction clientTransaction;
		ServerTransactionData(ClientTransaction ct) {
			clientTransaction = ct; 
		}
	}
	
	class ClientTransactionData {
		final static int MAX_RINGING = 10;
		int ringingCount = MAX_RINGING;
		ServerTransaction serverTransaction;
		ClientTransactionData(ServerTransaction st) {
			serverTransaction = st; 
		}
	}
	
	/**
	* Classe pour mmoriser les paramtres d'authentification 
	*  un serveur SMTP
	* Les ??? sont  modifier par des vrais paramtres
	*/
	protected class SMTPAuthenticator extends javax.mail.Authenticator
	{	
	    private static final String SMTP_AUTH_USER = "???";
		private static final String SMTP_AUTH_PWD = "???";

		public PasswordAuthentication getPasswordAuthentication()
	    {
	        String username = SMTP_AUTH_USER;
	        String password = SMTP_AUTH_PWD;
	        return new PasswordAuthentication(username, password);
	    }
	}

	/**
	 * Classe pour grer les temporisations relatives aux enregistrements des adresses SIP
	 */
	protected class RegisterTimerTask extends TimerTask {
		protected URI targetURI; 
		
		public RegisterTimerTask(URI uri) {
			targetURI = uri;
		}
		
		@Override
		public void run() {
			// Suppression de l'enregistrement aprs expiration du dlai
			registrations.remove(targetURI);
			deleteRegisterTimer(targetURI);
			logger.info("Registration time-out, removing:" + targetURI);
			
		}
		
	}
	
	
}
