package livresip.sipphone.rtp;

import java.io.IOException;

import javax.media.Controller;
import javax.media.ControllerClosedEvent;
import javax.media.ControllerEvent;
import javax.media.ControllerListener;
import javax.media.Manager;
import javax.media.MediaLocator;
import javax.media.NoProcessorException;
import javax.media.Processor;
import javax.media.control.BufferControl;
import javax.media.control.TrackControl;
import javax.media.format.AudioFormat;
import javax.media.protocol.ContentDescriptor;
import javax.media.protocol.DataSource;
import javax.media.rtp.RTPManager;
import javax.media.rtp.SendStream;

import org.apache.log4j.Logger;

/**
 * Classe grant l'emission d'un flux RTP depuis le priphrique de capture audio
 * 
 *  @see Chapitre 7 du livre Tlphonie SIP : concepts, usages et programmation en Java
 *	@author G. Vanwormhoudt
 */


public class AudioTransmiter implements ControllerListener {

	protected MediaLocator locator;
	protected Processor processor = null;
	protected RTPManager rtpManager;
	protected DataSource dataOutput = null;
	protected Logger logger = null;
	
	
	/**
	 * @param mngr manager  utiliser pour l'emission du flux
	 */
	public AudioTransmiter(RTPManager mngr, Logger lgr) {
		// priphrique de capture audio
		locator = new MediaLocator("javasound://8000");
		rtpManager = mngr;
		logger = lgr; 
	}
	
	public synchronized void start() throws Exception {

		// Cration du processeur de transcodage
		createProcessor();

		// ajout du flux au manager
		try {
			SendStream sendStream = rtpManager.createSendStream(dataOutput, 0);
			sendStream.start();

		} catch (Exception e) {
			processor.close();
			processor = null;
			throw e;
		}

		// Dmarre la transmission
		processor.start();
		logger.info("Starting RTP Session, sending data");
	}

	public void stop() {
		synchronized (this) {
			if (processor != null) {
				processor.stop();
				logger.info("Stopping RTP Session");
				processor.close();
				processor = null;
			}
		}
	}

	protected void createProcessor() throws Exception{
		if (locator == null)
			throw new Exception("Locator is null");


		// cration de la datasource
		DataSource ds;
		try {
			ds = Manager.createDataSource(locator);
		} catch (Exception e) {
			throw new Exception("Couldn't create DataSource");
		}
		
		// gestion du buffer
		BufferControl bc = (BufferControl)ds.getControl("javax.media.control.BufferControl");
		bc.setBufferLength(50);
		
		
		// cr le processeur en lui passant la datasource
		// 
		try {
			processor = Manager.createProcessor(ds);
		} catch (NoProcessorException npe) {
			throw new Exception("Couldn't create processor");
		} catch (IOException ioe) {
			throw new Exception("IOException creating processor");
		}

		// Attente de configuration
		if (!waitForState(processor, Processor.Configured)) {
			throw new Exception("Cannot configure processor");
		}

		// Rcupration des pistes du processeur
		TrackControl[] tracks = processor.getTrackControls();

		if (tracks == null || tracks.length < 1)
			throw new Exception("Couldn't find tracks in processor");

		// On fixe le format en PCM
		processor.setContentDescriptor(new ContentDescriptor(ContentDescriptor.RAW_RTP));

		boolean encodingOk = false;
		
		// Programmation des pistes
		// Boucle sur l'ensemble des pistes
		for (int i = 0; i < tracks.length; i++) {
			if (!encodingOk) {
				// on active la piste si on reussi  programmer la piste
				if (tracks[i].setFormat(new AudioFormat(AudioFormat.ULAW_RTP, 8000, 8, 1)) == null) {
					// on la desactive sinon
					tracks[i].setEnabled(false);
				}
				else {
					encodingOk = true;
				}
			}
			// on desactive toutes les autres pistes
			else {
				tracks[i].setEnabled(false);
			}
		}

		// attente du l'tat Realized
		if (!waitForState(processor, Controller.Realized)) {
			throw new Exception("Cannot realize processor");
		}

		// rcuperation de la sortie du processeur
		dataOutput = processor.getDataOutput();
	}

	protected Integer stateLock = new Integer(0);
	protected boolean failed = false;

	Integer getStateLock() {
		return stateLock;
	}

	void setFailed() {
		failed = true;
	}

	protected synchronized boolean waitForState(Processor p, int state) {
		p.addControllerListener(this);
		failed = false;

		// Call the required method on the processor
		if (state == Processor.Configured) {
			p.configure();
		} else if (state == Processor.Realized) {
			p.realize();
		}

		// Attente jusqu' obtention d'un vnement 
		// confirmant le succs ou l'chec de la ralisation
		while (p.getState() < state && !failed) {
			synchronized (getStateLock()) {
				try {
					getStateLock().wait();
				} catch (InterruptedException ie) {
					return false;
				}
			}
		}

		if (failed)
			return false;
		else
			return true;
	}

	@Override
	public void controllerUpdate(ControllerEvent ce) {
		// Si il y une error during la configuration ou
		// la ralisation, on ferme le processeur
		if (ce instanceof ControllerClosedEvent)
			setFailed();

		// Tout vnement de contrle se traduit 
		// par une notification au thread en attente dans waitForState
		if (ce instanceof ControllerEvent) {
			synchronized (getStateLock()) {
				getStateLock().notifyAll();
			}
		}
		
	}

}
