/*
 * Copyright (C) 2010 The Android Open Source Project
 *
 * Licensed under the Apache License, Version 2.0 (the "License");
 * you may not use this file except in compliance with the License.
 * You may obtain a copy of the License at
 *
 *      http://www.apache.org/licenses/LICENSE-2.0
 *
 * Unless required by applicable law or agreed to in writing, software
 * distributed under the License is distributed on an "AS IS" BASIS,
 * WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
 * See the License for the specific language governing permissions and
 * limitations under the License.
 */

package livresip.android;

import android.app.Activity;
import android.app.AlertDialog;
import android.app.Dialog;
import android.content.DialogInterface;
import android.content.Intent;
import android.content.SharedPreferences;
import android.net.sip.SipProfile;
import android.os.Bundle;
import android.preference.PreferenceManager;
import android.view.LayoutInflater;
import android.view.Menu;
import android.view.MenuItem;
import android.view.MotionEvent;
import android.view.View;
import android.view.WindowManager;
import android.widget.EditText;
import android.widget.TextView;
import android.widget.ToggleButton;

/**
 * Classe grant l'cran principal de l'application WalkieTalkie
 * 
 * @see Chapitre 8 du livre Tlphonie SIP : concepts, usages et programmation
 *      en Java
 * @author G. Vanwormhoudt
 */
public class WalkieTalkieActivity extends Activity implements
		View.OnTouchListener, WalkieTalkieSipListener {

	public static final String tag = WalkieTalkieActivity.class.getSimpleName();

	private static final int CALL_ADDRESS = 1;
	private static final int SET_AUTH_INFO = 2;
	private static final int UPDATE_SETTINGS_DIALOG = 3;
	private static final int HANG_UP = 4;

	protected WalkieTalkieSip walkieTalkieSip;

	protected TextView statusBox;

	@Override
	public void onCreate(Bundle savedInstanceState) {

		super.onCreate(savedInstanceState);
		setContentView(R.layout.walkietalkie);

		// gestion des composants d'interface
		ToggleButton pushToTalkButton = (ToggleButton) findViewById(R.id.pushToTalk);
		pushToTalkButton.setOnTouchListener(this);
		statusBox = (TextView) findViewById(R.id.sipLabel);

		// on maintient l'cran actif pour le "Push to Talk"
		getWindow().addFlags(WindowManager.LayoutParams.FLAG_KEEP_SCREEN_ON);

		// creation de la couche SIP
		walkieTalkieSip = new WalkieTalkieSip(this);
	}

	@Override
	public void onStart() {
		super.onStart();

		// cette mthode est  appele  la premiere apparition ou
		// quand on revient de l'cran des prfrences
		// on (re)fait donc la configuration SIP pour tenir compte des
		// prfrences
		setupSip();
	}

	@Override
	public void onDestroy() {
		super.onDestroy();

		// on termine proprement en quittant
		walkieTalkieSip.terminate();
	}

	public WalkieTalkieSip getWalkieTalkieSip() {
		return walkieTalkieSip;
	}

	/**
	 * Configuration de la couche SIP
	 */
	public void setupSip() {
		SharedPreferences prefs = PreferenceManager
				.getDefaultSharedPreferences(getBaseContext());
		String username = prefs.getString("namePref", "");
		String domain = prefs.getString("domainPref", "");
		String password = prefs.getString("passPref", "");
		String proxy = prefs.getString("proxyPref", "");

		if (username.length() == 0 || domain.length() == 0
				|| password.length() == 0 || proxy.length() == 0) {
			showDialog(UPDATE_SETTINGS_DIALOG);
			return;
		}
		updateStatus("");
		walkieTalkieSip.setupProfile(username, domain, password, proxy);
	}

	/**
	 * Gestion des messages d'information selon le type d'vnement
	 * 
	 * @param e
	 * @param data
	 */
	public void notify(Event e, Object data) {
		switch (e) {
		case CONNECTED:
			SipProfile peerProfile = (SipProfile) data;
			String useName = peerProfile.getDisplayName();
			if (useName == null) {
				useName = peerProfile.getUserName();
			}
			updateStatus(useName + "@" + peerProfile.getSipDomain());
			break;
		case DISCONNECTED:
			updateStatus("Ready.");
			break;
		case REGISTERED:
			updateStatus("Registered.");
			break;
		case REGISTERING:
			updateStatus("Registering with SIP Server...");
			break;
		case REGISTRATION_FAIL:
			updateStatus("Registration failed.  Please check settings.");
			break;
		}
	}

	/**
	 * Mise  jour de la zone affichant les messages d'informations
	 * 
	 * @param status
	 */
	public void updateStatus(final String status) {
		// Mise  jour de l'interface dans un thread apart
		this.runOnUiThread(new Runnable() {
			public void run() {
				TextView labelView = (TextView) findViewById(R.id.sipLabel);
				labelView.setText(status);
			}
		});
	}

	/**
	 * Updates whether or not the user's voice is muted, depending on whether
	 * the button is pressed.
	 * 
	 * @param v
	 *            The View where the touch event is being fired.
	 * @param event
	 *            The motion to act on.
	 * @return boolean Returns false to indicate that the parent view should
	 *         handle the touch event as it normally would.
	 */
	public boolean onTouch(View v, MotionEvent event) {
		if (!walkieTalkieSip.isConnected())
			return false;
		if (event.getAction() == MotionEvent.ACTION_DOWN
				&& walkieTalkieSip.isCallMuted()) {
			walkieTalkieSip.toggleMute();
		} else if (event.getAction() == MotionEvent.ACTION_UP
				&& !walkieTalkieSip.isCallMuted()) {
			walkieTalkieSip.toggleMute();
		}
		return false;
	}

	@Override
	public boolean onCreateOptionsMenu(Menu menu) {
		menu.add(0, CALL_ADDRESS, 0, "Call someone");
		menu.add(0, SET_AUTH_INFO, 0, "Edit your SIP Info.");
		menu.add(0, HANG_UP, 0, "End Current Call.");

		return true;
	}

	@Override
	public boolean onOptionsItemSelected(MenuItem item) {
		switch (item.getItemId()) {
		case CALL_ADDRESS:
			showDialog(CALL_ADDRESS);
			break;
		case SET_AUTH_INFO:
			updatePreferences();
			break;
		case HANG_UP:
			walkieTalkieSip.endCall();
			updateStatus("Ready.");
			break;
		}
		return true;
	}

	@Override
	protected Dialog onCreateDialog(int id) {
		switch (id) {
		case CALL_ADDRESS:

			LayoutInflater factory = LayoutInflater.from(this);
			final View textBoxView = factory.inflate(
					R.layout.call_address_dialog, null);
			return new AlertDialog.Builder(this).setTitle("Call Someone.")
					.setView(textBoxView).setPositiveButton(
							android.R.string.ok,
							new DialogInterface.OnClickListener() {
								public void onClick(DialogInterface dialog,
										int whichButton) {
									EditText textField = (EditText) (textBoxView
											.findViewById(R.id.calladdress_edit));
									String sipAddress = textField.getText()
											.toString();
									updateStatus(sipAddress);
									walkieTalkieSip.initiateCall(sipAddress);

								}
							}).setNegativeButton(android.R.string.cancel,
							new DialogInterface.OnClickListener() {
								public void onClick(DialogInterface dialog,
										int whichButton) {
									// Noop.
								}
							}).create();

		case UPDATE_SETTINGS_DIALOG:
			return new AlertDialog.Builder(this).setMessage(
					"Please update your SIP Account Settings.")
					.setPositiveButton(android.R.string.ok,
							new DialogInterface.OnClickListener() {
								public void onClick(DialogInterface dialog,
										int whichButton) {
									updatePreferences();
								}
							}).setNegativeButton(android.R.string.cancel,
							new DialogInterface.OnClickListener() {
								public void onClick(DialogInterface dialog,
										int whichButton) {
									// Noop.
								}
							}).create();
		}
		return null;
	}

	public void updatePreferences() {
		Intent settingsActivity = new Intent(getBaseContext(),
				SipSettings.class);
		startActivity(settingsActivity);
	}

}
