package livresip.android;

import livresip.android.WalkieTalkieSipListener.Event;
import android.app.PendingIntent;
import android.content.Context;
import android.content.Intent;
import android.content.IntentFilter;
import android.net.sip.SipAudioCall;
import android.net.sip.SipException;
import android.net.sip.SipManager;
import android.net.sip.SipProfile;
import android.net.sip.SipRegistrationListener;
import android.util.Log;

/**
 * Classe grant la signalisation SIP de l'application WalkieTalkie
 * 
 * @see Chapitre 8 du livre Tlphonie SIP : concepts, usages et programmation
 *      en Java
 * @author G. Vanwormhoudt
 */

public class WalkieTalkieSip extends SipAudioCall.Listener implements
		SipRegistrationListener {

	public static final String tag = WalkieTalkieActivity.class.getSimpleName();

	protected SipManager manager = null;
	protected SipProfile profile = null;
	protected SipAudioCall currentCall = null;
	protected IncomingCallReceiver callReceiver;

	protected Context context;
	protected WalkieTalkieSipListener listener;

	public WalkieTalkieSip(Context c) {
		context = c;
		listener = (WalkieTalkieSipListener) c;

		initializeReceiver();
		initializeManager();
	}

	protected void initializeReceiver() {
		// On commence par enregistrer le broadcast receiver qui va recevoir les
		// appels
		// L'intent est configur via la mthode open du SipManager (liaison
		// avec un profile)

		IntentFilter filter = new IntentFilter();
		filter.addAction("android.SipDemo.INCOMING_CALL");
		callReceiver = new IncomingCallReceiver();
		context.registerReceiver(callReceiver, filter);
	}

	public void initializeManager() {
		if (manager == null) {
			manager = SipManager.newInstance(context);
		}
	}

	/**
	 * Enregistre le compte SIP dans un proxy
	 * 
	 * @param username
	 * @param domain
	 * @param password
	 * @param proxy
	 */
	public void setupProfile(String username, String domain, String password,
			String proxy) {
		if (manager == null) {
			return;
		}

		if (profile != null) {
			closeLocalProfile();
		}

		try {
			SipProfile.Builder builder = new SipProfile.Builder(username,
					domain);
			if (!password.equals(""))
				builder.setPassword(password);
			if (!proxy.equals(""))
				builder.setOutboundProxy(proxy);
			builder.setSendKeepAlive(false);
			builder.setDisplayName(username);
			profile = builder.build();

			Intent i = new Intent();
			i.setAction("android.SipDemo.INCOMING_CALL");
			PendingIntent pi = PendingIntent.getBroadcast(context, 0, i,
					Intent.FILL_IN_DATA);
			manager.open(profile, pi, null);

			// LՎcouteur est ajout aprs louverture du profile pour
			// tre sr que la notification se fasse
			manager.setRegistrationListener(profile.getUriString(), this);

		} catch (Exception pe) {
			Log.d(tag, "Error during profile setup");
		}
	}

	/**
	 * Closes out your local profile, freeing associated objects into memory and
	 * unregistering your device from the server.
	 */
	public void closeLocalProfile() {
		if (manager == null) {
			return;
		}
		try {
			if (profile != null) {
				manager.close(profile.getUriString());
			}
		} catch (Exception ee) {
			Log.d("WalkieTalkieActivity/onDestroy",
					"Failed to close local profile.", ee);
		}
	}

	/**
	 * Gre la fermeture de la couche SIP
	 */
	public void terminate() {
		if (currentCall != null) {
			currentCall.close();
		}
		closeLocalProfile();

		if (callReceiver != null) {
			context.unregisterReceiver(callReceiver);
		}
	}

	/**
	 * Dtermine si l'appel est tabli
	 * 
	 * @return
	 */
	public boolean isConnected() {
		return currentCall != null;
	}

	/**
	 * Etablit un appel avec une addresse SIP donne
	 * 
	 * @param sipAddress
	 */
	public void initiateCall(String sipAddress) {
		try {
			if (!isConnected()) {
				currentCall = manager.makeAudioCall(profile.getUriString(),
						sipAddress, this, 30);
			}
		} catch (Exception e) {
			Log.i(tag, "Error when trying to initiateCall.", e);
			if (currentCall != null) {
				currentCall.close();
			}
		}
	}

	/**
	 * Met fin  un appel
	 */
	public void endCall() {
		if (isConnected()) {
			try {
				currentCall.endCall();
			} catch (SipException se) {
				Log.d(tag, "Error ending call.", se);
			}
			currentCall.close();
		}
	}

	/**
	 * Gre la rception d'un appel
	 * 
	 * @param intent
	 */
	public void takeAudioCall(Intent intent) {
		try {
			// manager.takeAudioCall(intent, this);

			SipAudioCall incomingCall = manager.takeAudioCall(intent, this);
			incomingCall.answerCall(30);
			incomingCall.startAudio();
			incomingCall.setSpeakerMode(true);
			if (incomingCall.isMuted()) {
				incomingCall.toggleMute();
			}

		} catch (Exception e) {
			Log.d(tag, "Error taking Call");
		}

	}

	/**
	 * Active le mode silencieux
	 */

	public void toggleMute() {
		if (currentCall != null) {
			currentCall.toggleMute();
		}
	}

	/**
	 * Dtermine si le mode silencieux est actif
	 * 
	 * @return
	 */
	public boolean isCallMuted() {
		if (isConnected()) {
			return currentCall.isMuted();
		}
		return false;
	}

	public boolean isCurrentCall(SipAudioCall call) {
		return (currentCall == call);
	}

	@Override
	public void onCallEstablished(SipAudioCall call) {
		if (!isConnected()) {
			currentCall = call;
			currentCall.startAudio();
			currentCall.setSpeakerMode(true);
			if (currentCall.isMuted()) {
				currentCall.toggleMute();
			}
			listener.notify(Event.CONNECTED, call.getPeerProfile());
		} else {
			closeCall(call);
		}
	}

	@Override
	public void onCallBusy(SipAudioCall call) {
		if (isCurrentCall(call)) {
			listener.notify(Event.BUSY, call.getPeerProfile());
			currentCall = null;
		}
		closeCall(call);
	}

	@Override
	public void onCallEnded(SipAudioCall call) {
		if (isCurrentCall(call)) {
			listener.notify(Event.DISCONNECTED, call);
			currentCall = null;
		}
		closeCall(call);
	}

	protected void closeCall(SipAudioCall call) {
		try {
			call.endCall();
		} catch (SipException e) {
			e.printStackTrace();
		} finally {
			call.close();
		}
	}

	@Override
	public void onRinging(SipAudioCall call, SipProfile caller) {
		try {
			if (!isConnected()) {
				call.answerCall(30);
			} else {
				closeCall(call);
			}
		} catch (Exception e) {
			e.printStackTrace();
		}
	}

	public void onRegistering(String localProfileUri) {
		listener.notify(Event.REGISTERING, localProfileUri);
	}

	public void onRegistrationDone(String localProfileUri, long expiryTime) {
		listener.notify(Event.REGISTERED, localProfileUri);
	}

	public void onRegistrationFailed(String localProfileUri, int errorCode,
			String errorMessage) {
		listener.notify(Event.REGISTRATION_FAIL, localProfileUri);
	}

}
