package livresip.httpservlet;
import java.io.IOException;
import java.io.PrintWriter;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Set;
import java.util.Map.Entry;

import javax.servlet.ServletConfig;
import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.http.HttpServlet;
import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import javax.servlet.sip.Address;
import javax.servlet.sip.SipApplicationSession;
import javax.servlet.sip.SipFactory;
import javax.servlet.sip.SipServlet;
import javax.servlet.sip.SipServletRequest;
import javax.servlet.sip.SipSession;
import javax.servlet.sip.SipURI;

import livresip.sipservlet.CallInfo;
import livresip.sipservlet.ProxySipServlet;

/**
 * Classe de Servlet HTTP pour afficher les adresses SIP enregistrs et leur envoyer un message texte
 * 
 * @see Chapitre 8 du livre Tlphonie SIP : concepts, usages et programmation
 *      en Java
 * @author G. Vanwormhoudt
 */

public class ListUsersHttpServlet extends HttpServlet {

    public static final String LASTMESSAGE_MAP = "LASTMESSAGE_MAP";
    protected HashMap<SipURI, String> messagesMap;

    protected SipFactory sipFactory;
    protected ServletContext ctx;


   @Override
   public void init(ServletConfig config) throws ServletException {
     ctx = config.getServletContext();
     sipFactory = (SipFactory) ctx.getAttribute(SipServlet.SIP_FACTORY);
     if ( sipFactory == null ) {
            throw new ServletException("Factory not initialized");
     }
     messagesMap = new HashMap<SipURI, String>();
     ctx.setAttribute(LASTMESSAGE_MAP, messagesMap);
   }


    @Override
    protected void doGet(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        response.setContentType("text/html;charset=UTF-8");
        PrintWriter out = response.getWriter();

        try {
            generateHtml(out);
        } finally {
            out.close();
        }
    }

    protected void generateHtml(PrintWriter out) {
        HashMap<SipURI, SipURI> registrations = getRegistrationsMap();
        out.println("<html>");
        out.println("<head>");
        out.println("<title>Servlet ListRegisteredUsers</title>");
        out.println("</head>");
        out.println("<body>");
        out.println("<h1>List of registered users</h1>");
        out.println("<hr>");

        if (registrations == null) {
            out.println("<h2>None</h2>");
        }
        else {
            out.println("<form action='ListUsers' method='GET'>");
            out.println("<table border='1'>");
            out.println("<tr><th>Registered users</th><th>Last caller</th><th>Date</th><th>Last Message</th></tr>");
            SipURI key;
            CallInfo lastCall = null;
            String msg = null;
            HashMap<SipURI, CallInfo> callerMap = getLastCallerMap();

            for (Entry<SipURI, SipURI>  entry : registrations.entrySet()) {
                key = entry.getKey();
                if (callerMap != null) {
                    lastCall = callerMap.get(key);
                }
                out.println("<tr>");
                out.println("<td>");
                out.println("<input type='checkbox' name='contact' value='" + key + "'>" + key);
                out.println("</td>");
                out.println("<td>");
                if (lastCall != null) {
                    out.println(lastCall.getCaller());
                } else {
                    out.println("-");
                }
                out.println("</td>");
                out.println("<td>");
                if (lastCall != null) {
                    out.println(lastCall.getDate());
                } else {
                    out.println("-");
                }
                out.println("</td>");
                out.println("<td>");
                msg = messagesMap.get(key);
                if (msg != null) {
                    int len = msg.length();
                    if (len > 20) msg = msg.substring(0, 20) + "...";
                    out.println(msg);
                } else {
                    out.println("None");
                }

                out.println("</td>");
                out.println("</tr>");
            }
            out.println("</table>");
            out.println("<input type='submit' value='Refresh' onClick='this.form.method=\'GET\''/>");
            out.println("<hr>");
            out.println("Send a message to checked users<br>");
            out.println("<input type='text' name='msg'/><br>");
            out.println("<input type='submit' value='Send' onClick='this.form.method=\'POST\''/>");
            out.println("</form>");

            cleanMessagesMap();
        }
        out.println("</body>");
        out.println("</html>");


    }


    @Override
    protected void doPost(HttpServletRequest request, HttpServletResponse response)
    throws ServletException, IOException {
        String[] contacts = request.getParameterValues("contact");
        String msg = request.getParameter("msg");

        if ( contacts != null && msg != null ) {
            SipApplicationSession appSession = sipFactory.createApplicationSession();
            for (String contact: contacts) {
                SipURI contactURI = (SipURI) sipFactory.createURI(contact);
                SipURI toURI = getRegistrationsMap().get(contactURI);
                Address to = sipFactory.createAddress(toURI);
                Address from = sipFactory.createAddress("sip:convergedapp@domain.org");
    
                SipServletRequest sipReq = (SipServletRequest) sipFactory.createRequest(appSession, "MESSAGE", from, to);
                sipReq.setContent(msg, "text/plain");

                // set servlet to invoke by reponse
                SipSession session = sipReq.getSession();
                session.setAttribute("lastmessage", msg);
                session.setAttribute("contact", contactURI);

                session.setHandler("messagehandler");

                sipReq.send();    
            }
        }
    
        response.setContentType("text/html;charset=UTF-8");
        PrintWriter out = response.getWriter();
        try {
            generateHtml(out);
        } finally {
            out.close();
        }
    }

    protected HashMap<SipURI, SipURI> getRegistrationsMap() {
        return  (HashMap<SipURI, SipURI>) ctx.getAttribute(ProxySipServlet.REGISTRATION_MAP);
    }

   protected HashMap<SipURI, CallInfo> getLastCallerMap() {
        return  (HashMap<SipURI, CallInfo>) ctx.getAttribute(ProxySipServlet.LASTCALLER_MAP);
   }

    private void cleanMessagesMap() {
        Set<SipURI> uris = new HashSet<SipURI>();
        for (SipURI uri: messagesMap.keySet()) {
            if (! getRegistrationsMap().containsKey(uri)) {
                uris.add(uri);
            }
        }
        for (SipURI uri: uris) {
            messagesMap.remove(uri);
        }

    }
}
