package livresip.sipservlet;
import java.io.IOException;
import java.util.Date;
import java.util.HashMap;
import java.util.Iterator;

import javax.servlet.ServletContext;
import javax.servlet.ServletException;
import javax.servlet.sip.Address;
import javax.servlet.sip.Proxy;
import javax.servlet.sip.ServletTimer;
import javax.servlet.sip.SipApplicationSession;
import javax.servlet.sip.SipFactory;
import javax.servlet.sip.SipServlet;
import javax.servlet.sip.SipServletRequest;
import javax.servlet.sip.SipServletResponse;
import javax.servlet.sip.SipSession;
import javax.servlet.sip.SipURI;
import javax.servlet.sip.TimerListener;
import javax.servlet.sip.TimerService;



/**
 * Classe principale de Servlet SIP implantant le miniproxy
 * 
 * @see Chapitre 8 du livre Tlphonie SIP : concepts, usages et programmation
 *      en Java
 * @author G. Vanwormhoudt
 */

public class ProxySipServlet extends SipServlet
        implements TimerListener {

 
	private static final long serialVersionUID = 1L;
	
	public static final String REGISTRATION_MAP = "REGISTRATION_MAP";
    public static final String LASTCALLER_MAP = "LASTCALLER_MAP";
    
    protected HashMap<SipURI, SipURI> registrationsMap;
    protected HashMap<SipURI, CallInfo> callersMap;
    protected HashMap<SipURI, ServletTimer> timersMap;

    protected ServletContext ctx;

    protected SipFactory    sipFactory;
    protected TimerService  timerService;
    
    @Override
    public void init(javax.servlet.ServletConfig config) throws ServletException {
        super.init(config);
        ctx = config.getServletContext();

        sipFactory = (SipFactory) ctx.getAttribute(SIP_FACTORY);
        timerService = (TimerService) ctx.getAttribute(TIMER_SERVICE);
        if ( sipFactory == null || timerService == null ) {
            throw new ServletException("Services not initialized");
        }
        timersMap = new HashMap<SipURI, ServletTimer>();
        registrationsMap = new HashMap<SipURI, SipURI>();
        ctx.setAttribute(REGISTRATION_MAP, registrationsMap);

        callersMap = new HashMap<SipURI, CallInfo>();
        ctx.setAttribute(LASTCALLER_MAP, callersMap);
    }
   
    protected void doRegister(SipServletRequest request) throws ServletException, IOException {

        SipServletResponse response = request.createResponse(200);
        try {
            SipURI to = cleanURI((SipURI) request.getTo().getURI());
            Address na = request.getAddressHeader("Contact");
            SipURI contact = (SipURI) na.getURI();
            
            // Extraction et vrification du paramtre ou du champ Expires
            String expiresString = na.getParameter("expires");
            if (expiresString == null) { // Check the Expires header
                expiresString = request.getHeader("Expires");
            }
            
            // Si la valeur n'existe pas, on en prend une par dfault
            if (expiresString == null) {
                expiresString = "3600";
                response.setExpires(3600);
            }
            long expires = Integer.parseInt(expiresString);
            
            // Annulation d'un enregistrement
            if (expires == 0) {
                registrationsMap.remove(to);
                callersMap.remove(to);

                // Arrt du temporisateur
                ServletTimer st = timersMap.remove(to);
                if (st != null) {
                    st.cancel();
                }
            } else {
                // Insertion ou modification d'un enregistrement
            	registrationsMap.put(to, contact);
                if (timerService != null) {
                	// Creation d'un nouveau temporisateur
                    ServletTimer st = timerService.createTimer(
                                                    request.getSession().getApplicationSession(),
                                                    1000 * expires,
                                                    false,
                                                    to.toString());

                     st = timersMap.put(to, st);
                     // Arrt du temporisateur prcdent
                     if (st != null) st.cancel();
                }
                response.setHeader("Contact",na.toString());
            }
            // Envoi de la rponse
            response.send();
        } catch(Exception e) {
            response.setStatus(500);
            response.send();
        }

        SipApplicationSession appsess = request.getApplicationSession(false);
        if (appsess != null) {
            appsess.invalidate();
        }
    }

    protected void doInvite(SipServletRequest request) throws ServletException, IOException {
        SipURI toURI = cleanURI((SipURI) request.getTo().getURI());
        SipURI target = registrationsMap.get(toURI);

        if (target != null) {
                Proxy proxy = request.getProxy();
                proxy.setRecordRoute(false);
                proxy.setSupervised(true);
                proxy.proxyTo(target);
        }
        else {
            SipServletResponse response = request.createResponse(404);
            response.send();
        }
    }


    protected void doSuccessResponse(SipServletResponse resp)
                          throws javax.servlet.ServletException,
                                 java.io.IOException {
        SipSession session = resp.getSession();

        SipURI fromUri = cleanURI((SipURI) session.getRemoteParty().getURI());
        SipURI toUri = cleanURI((SipURI) session.getLocalParty().getURI());


        callersMap.put(toUri, new CallInfo(fromUri, new Date()));
        session.invalidate();
    }

    protected SipURI cleanURI(SipURI original) {
        SipURI copy = (SipURI) original.clone();
        Iterator headers = copy.getHeaderNames();
        if (headers.hasNext()) {
            headers.next();
            headers.remove();
        }
        Iterator parameters = copy.getParameterNames();
        if (parameters.hasNext()) {
            String param = (String) parameters.next();
            copy.removeParameter(param);
        }
        return copy;
    }

    public void timeout(ServletTimer st) {
        try {
            SipURI uri = (SipURI) sipFactory.createURI((String) st.getInfo());
            registrationsMap.remove(uri);
            callersMap.remove(uri);
            timersMap.remove(st);
        }
        catch(Exception e) {
            e.printStackTrace();
        }
    }

  
}

